/*
 * $Id: 80211.c,v 1.5 2005/03/09 09:22:28 kos Exp $
 * 
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <string.h>

#include "ssidsniff.h"
#include "llog.h"
#include "util.h"
#include "80211.h"

static int
f80211_parse_tags(unsigned char *p, unsigned int l, f80211_t *res) {
  unsigned int le;
  unsigned int tlen, ttype;

  for(le=0; le <= l; ) {
      
    if(le+1 >= l)
      break;

    ttype=p[le];
    tlen=p[le+1];

    if(le+1+tlen > l) {
      llog(6, "parse: problem decoding tag, ignoring\n");
      break;
    }

    switch(ttype) {
    case TAG_SSID:
      llog(4, "   tag SSID: ");
      memset(res->ssid, '\0', DEF_LINELEN);
      memcpy(res->ssid, &p[le+2], tlen);
      llog(4, "%s\n", res->ssid);
      break;
    case TAG_SUPP_RATES:
      llog(4, "   tag RATES: \n");
      break;
    case TAG_DS_PARAMETER:
      llog(4, "   tag DSPARAMS: ");
      res->channel = p[le+2];
      llog(4, "channel %d\n", res->channel);
      break;
    case TAG_IBSS_PARAMETER:
      llog(4, "   tag IBSSPARAMS:\n");
      break;
    default:
      llog(4, "   unhandled tag %u\n", ttype);
      break;
    }

    le += tlen + 2;
  }
  
  return 0;
}

int
f80211_parse(unsigned char *p, unsigned int l, f80211_t *res) {
  f80211_mac_t *mac;
  frame_control_t *fc;
  beacon_header_t *bh;

  if(l<sizeof(f80211_mac_t)) {
    llog(6, "parse: packet too small to have 802.11 MAC header\n");
    return 0;
  }

  mac = (f80211_mac_t *)p;
  memcpy(&res->f_mac, mac, sizeof(f80211_mac_t));
  fc = (frame_control_t *)mac->fc;
  if(fc->version) {
    llog(6, "parse: unhandled version number 0x%x\n", fc->version);
    return 0;
  }

  llog(4, "MAC 80211\n FC: version %u, type %u, subtype %u\n", 
       fc->version, fc->type, fc->subtype);

  res->type = fc->type;
  res->subtype = fc->subtype;
  res->wepused = fc->wep;   /* wepused for _this_ packet */
  strncpy(res->mac, mac2ascii(mac->address2), DEF_LINELEN);
  strncpy(res->bss, mac2ascii(mac->address3), DEF_LINELEN);
  strncpy(res->dest, mac2ascii(mac->address1), DEF_LINELEN);
  
  switch(fc->subtype) {
  case MGT_PROBE_REQ:
    llog(4, " Type: Probe Request\n");
    llog(4, " Destination: %s\n", res->dest);
    llog(4, " Source: %s\n", res->mac);
    llog(4, " BSS: %s\n", res->bss);
    /* tagged parameters */
    f80211_parse_tags(p + sizeof(f80211_mac_t),
		      l - sizeof(f80211_mac_t),
		      res);
    break;

  case MGT_BEACON:
    if(sizeof(beacon_header_t)+sizeof(f80211_mac_t) > l) {
      llog(4, "parse: bad frame size\n");
      return 0;
    }

    bh = (beacon_header_t *)(p+sizeof(f80211_mac_t));
    llog(4, " Type: Beacon\n");
    llog(4, " Destination: %s\n", res->dest);
    llog(4, " Source: %s\n", res->mac);
    llog(4, " BSS: %s\n", res->bss);
    llog(4, " Sequence: %02x%02x\n", mac->seq[1], mac->seq[0]);
    llog(4, "  beacon timestamp: %s\n", hex2ascii(bh->timestamp, 8));
    llog(4, "  beacon interval: %s\n", hex2ascii(bh->interval, 2));
    llog(4, "  beacon capability: %s\n", hex2ascii(bh->capability, 2));
    
    /* fixed parameters */
    res->ap  = bh->capability[0] & IEEE80211_CAPINFO_ESS;
    res->wepcapable = bh->capability[0] & IEEE80211_CAPINFO_PRIVACY;
    
    /* tagged parameters */
    f80211_parse_tags(p + (sizeof(f80211_mac_t)+sizeof(beacon_header_t)), 
		      l - (sizeof(f80211_mac_t)+sizeof(beacon_header_t)),
		      res);
    break;
  default:
    llog(6, "parse: unhandled packet subtype %d\n", *p>>4);
    break;
  } 

  return 0;
}
