/*
 * $Id: mach-linux.c,v 1.5 2005/05/30 04:18:19 kos Exp $
 *
 * Some bits taken from wireless-tools-27
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

// #include <net/if.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/ioctl.h>
#include <linux/wireless.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>

#include "mach.h"
#include "util.h"

#define KILO    1e3
#define MEGA    1e6
#define GIGA    1e9

static double
iw_freq2float(const struct iw_freq *in) {
  int           i;
  double        res = (double) in->m;
  
  for(i = 0; i < in->e; i++)
    res *= 10;
  return(res);
}

int
iw_channel_to_freq(int                          channel,
                   double *                     pfreq,
                   const struct iw_range *      range)
{
  int           has_freq = 0;
  int           k;

  /* Check if the driver support only channels or if it has frequencies */
  for(k = 0; k < range->num_frequency; k++)
    {
      if((range->freq[k].e != 0) || (range->freq[k].m > (int) KILO))
        has_freq = 1;
    }
  if(!has_freq)
    return(-1);

  /* Find the correct frequency in the list */
  for(k = 0; k < range->num_frequency; k++)
    {
      if(range->freq[k].i == channel)
        {
          *pfreq = iw_freq2float(&(range->freq[k]));
          return(channel);
        }
    }
  /* Not found */
  return(-2);
}

int
iw_freq_to_channel(double                       freq,
                   const struct iw_range *      range)
{
  double        ref_freq;
  int           k;

  /* Check if it's a frequency or not already a channel */
  if(freq < KILO)
    return(-1);

  /* We compare the frequencies as double to ignore differences
   * in encoding. Slower, but safer... */
  for(k = 0; k < range->num_frequency; k++)
    {
      ref_freq = iw_freq2float(&(range->freq[k]));
      if(freq == ref_freq)
        return(range->freq[k].i);
    }
  /* Not found */
  return(-2);
}

double
mach_freq_get(char *device) {
  int skfd;
  struct iwreq wrq;
  double res;
  char buffer[sizeof(struct iw_range)*2];
  struct iw_range *range;
  int channel;
  
  skfd = socket(AF_INET, SOCK_DGRAM, 0);
  strcpy(wrq.ifr_name, device);
  if(ioctl(skfd, SIOCGIWFREQ, &wrq) < 0)
    res = -1;
  else 
    res = iw_freq2float(&wrq.u.freq);

  /* ewww */
  if(res < 1024) {
    channel = (int)res;
    wrq.u.data.pointer = (caddr_t) buffer;
    wrq.u.data.length = sizeof(buffer);
    wrq.u.data.flags = 0;
    strcpy(wrq.ifr_name, device);
    ioctl(skfd, SIOCGIWRANGE, &wrq);
    range = (struct iw_range *) buffer;
    
    iw_channel_to_freq(channel, &res, range);
  }

  close(skfd);
  return res;
}

int
mach_iswifi(char *device, char *descr, int len) {
  int skfd;
  struct iwreq wrq;
  int res;

  skfd = socket(AF_INET, SOCK_DGRAM, 0);
  strcpy(wrq.ifr_name, device);
  res = ioctl(skfd, SIOCGIWNAME, &wrq) < 0 ? 0 : 1;
  if(descr)
    strncpy(descr, "Unknown", len); // XXX
  close(skfd);
  return res;
}

int
mach_channel_set(char *device, int channel) {
  int skfd;
  struct iwreq wrq;
  int res;
  
  memset((void *)&wrq, '\0', sizeof(struct iwreq));
  strcpy(wrq.ifr_name, device);
  wrq.u.freq.m = channel;
  skfd = socket(AF_INET, SOCK_DGRAM, 0);
  res = ioctl(skfd, SIOCSIWFREQ, &wrq) < 0 ? 0 : 1;
  close(skfd);
  return res;  
}

int
mach_snr(char *device) {
  unsigned char line[MACH_LINELEN];
  FILE *fp;
  char *tokens[16];
  int ntokens;  
  char ldev[MACH_LINELEN];
  int snr;
  
  snprintf(ldev, MACH_LINELEN, "%s:", device);
  snr = 0;
  
  if(!(fp = fopen("/proc/net/wireless", "r")))
    return 1;
  
  fgets(line, MACH_LINELEN, fp); /* header */
  fgets(line, MACH_LINELEN, fp);
  while(fgets(line, MACH_LINELEN, fp)) {
    tokenize(line, tokens, &ntokens);
    if(!strcmp(ldev, tokens[0])) {
      sscanf(tokens[2], "%d.", &snr);
      fclose(fp);
      return snr;
    }
  }
  fclose(fp);
  return -1;
}

char *
mach_battery_status(void) {
  static unsigned char line[MACH_LINELEN];
  FILE *fp;
  char *tokens[16];
  int ntokens;  
  
  if(!(fp = fopen("/proc/apm", "r"))) {    
    return "None";
  }

  if(!fgets(line, MACH_LINELEN, fp)) {
    fclose(fp);
    return "None";
  }
  fclose(fp);

  tokenize(line, tokens, &ntokens);
  if(ntokens < 8) {
    return "None";
  }
  
  return tokens[6];
}

int
mach_set_if(char *device, int state) {
  int skfd;
  struct ifreq ifr;
  int res;

  skfd = socket(AF_INET, SOCK_DGRAM, 0); 
  strcpy(ifr.ifr_name, device);
  res = ioctl(skfd, SIOCGIFFLAGS, &ifr) < 0 ? 0 : 1;
  ifr.ifr_flags |= IFF_UP;
  res += ioctl(skfd, SIOCSIFFLAGS, &ifr) < 0 ? 0 : 1;  
  close(skfd);
  return res;
}

int
mach_set_monitor(char *device, int state) {
  int skfd;
  struct iwreq wrq;
  int res;

  skfd = socket(AF_INET, SOCK_DGRAM, 0); 
  strcpy(wrq.ifr_name, device);
  res = ioctl(skfd, SIOCGIWMODE, &wrq) < 0 ? 0 : 1;
  if(state)
    wrq.u.mode = IW_MODE_MONITOR;
  else
    wrq.u.mode = IW_MODE_INFRA;
  res += ioctl(skfd, SIOCSIWMODE, &wrq) < 0 ? 0 : 1;  
  close(skfd);
  return res;
}
