/*
 * $Id: ssidsniff.c,v 1.36 2005/05/30 04:18:19 kos Exp $
 *
 */

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <sys/types.h>
#include <sys/wait.h>
#include <sys/time.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <string.h>
#include <signal.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <errno.h>
#include <assert.h>

#if HAVE_GETOPT_H
#include <getopt.h>
#endif

#include "xmalloc.h"
#include "llog.h"
#include "ouidb.h"
#include "xpaudio.h"
#include "ssidsniff.h"
#include "util.h"
#include "80211.h"
#include "mach.h"
#include "cursesui.h"

#ifdef HAVE_PCAP_H
#include "pcap.h"
#endif

#ifdef HAVE_NET_BPF_H
#include "net/bpf.h"
#endif

ssidsniff_t *gs;
int pps_update_all(ssidsniff_t *s);

RETSIGTYPE
signal_handler(int sig) {
  switch(sig){
  case SIGINT:
    gs->running = 0;
    break;
  case SIGALRM:
    alarm(gs->delay);
    pps_update_all(gs);
    /* fall through */
  case SIGWINCH:
  case SIGCONT:
    ui_redraw(gs);    
    break;
  }
  signal(sig, signal_handler);
}

int
wnet_clear(wnet_t *res) {
  memset((void *)res, '\0', sizeof(wnet_t));
  return 0;
}

int
wnet_exists(ssidsniff_t *s, char *key) {
  int i;
  wnet_t *n;

  for(i=0; i<s->num_networks; i++) {
    n = &s->networks[i];
    if(!strcmp(n->frame.mac, key))
      return i;
  }
  return -1;
}

int
wnet_add(ssidsniff_t *s, char *key, void *b, unsigned int l) {
  wnet_t *n;
  void *p;

  if(s->num_networks == s->max_networks) {
    if(!(p = xrealloc(s->networks, (s->max_networks + DEF_WNET_ALLOC) * sizeof(wnet_t))))
      return 1;
    s->max_networks += DEF_WNET_ALLOC;
    s->networks = (wnet_t *)p;
  }

  n = &s->networks[s->num_networks];
  memcpy((void *)n, b, l);
  s->num_networks++;
  return 0;
}

int
wnet_get(ssidsniff_t *s, int which, void **b, unsigned int *l) {
  int i;
  
  for(i=0; i<s->num_networks; i++) {
    if(s->networks[i].serial == which) {
      *b = &s->networks[i];
      *l = sizeof(wnet_t);
      return 0;
    }
  }
  return 1;
}

int
counter_clear(counter_t *count) {
  count->total = count->beacon = count->data = count->martian = 0;
  return 0;
}

/* lazy version of pcap_dump */
int
save_packet(ssidsniff_t *s, struct pcap_pkthdr *h, u_char *sp) {
    
  if(s->dumpfile[0] && !s->pd_dump) {
    if(!(s->pd_dump = pcap_dump_open(s->pd, s->dumpfile))) {
      ui_info("pcap_dump_open(%s): %s", s->dumpfile, pcap_geterr(s->pd));
      s->dumpfile[0] = '\0';
      return 1;
    }
    ui_info("Capture file is %s", s->dumpfile);
  }

  pcap_dump((u_char *)s->pd_dump, h, sp);
  return 0;
}

/* make noise */
int
geiger_packet(ssidsniff_t *s, struct pcap_pkthdr *h, u_char *sp) {
  
  if(s->geiger && !s->xpa) {
    if(!(s->xpa = xpaudio_new(NULL))) {
      ui_info("Failed to start geiger mode");
      s->geiger = 0;
      return 1;
    }
    ui_info("Geiger mode on: %s", xpaudio_describe(s->xpa));
  }

  if(s->xpa)
    xpaudio_send(s->xpa, sp, h->len>DEF_GEIGER_L?DEF_GEIGER_L:h->len);    
  return 0;
}

double
timeval_diff(struct timeval t1, struct timeval t2) {
  double diff;
  
  diff = ((double)t2.tv_usec - (double)t1.tv_usec)/1000000 + 
    ((double)t2.tv_sec - (double)t1.tv_sec);
  return fabs(diff);
}

int
pps_update_all(ssidsniff_t *s) {
  struct timeval now;
  struct timezone tz;
  int i;
  wnet_t *n;
  double pps_cur;

  gettimeofday(&now, &tz);
  for(i=0; i<s->num_networks; i++) {
    n = &s->networks[i];
    pps_cur = (double)n->pps_delta / (double)timeval_diff(n->pps_prevts, now);
    n->pps = (int)(((double)n->pps + (double)pps_cur) / 2);
    //n->pps = 9*n->pps / 10;
    n->pps_prevts = now;
    n->pps_delta = 0;
  }
  return 0;
}

int
pps_update_net(wnet_t *n) {

  n->pps_delta++;
  return 0;
}

int
process_packet(ssidsniff_t *s) {
  struct pcap_pkthdr pkthdr;
  unsigned char *pkt;
  wnet_t sample;
  wnet_t *network;
  int serial;
  unsigned int l;

  if(!(pkt = (unsigned char *)pcap_next(s->pd, &pkthdr))) {
    ui_info("pcap_next: %s\n", pcap_geterr(s->pd));
    return 1;
  }
  
  if(s->capture == CAPTURE_ALL) 
    save_packet(s, &pkthdr, pkt);
    
  if(pkthdr.caplen != pkthdr.len) {
    ui_info("packet truncated, discarding\n");;
    return 2;
  }
    
  s->globalcount.total++;
  wnet_clear(&sample);
  f80211_parse(pkt, pkthdr.len, &sample.frame);
    
  /* if this packet doesn't have a MAC address, discard. Probably LLC */
  if(sample.frame.mac[0] == 0) {
    s->globalcount.martian++;
    return 3;
  }

  /* if this is a new network, create an entry */
  if((serial = wnet_exists(s, sample.frame.mac)) == -1) {
    sample.serial = s->num_networks;
    counter_clear(&sample.count);
    ouidb_get_vendor(sample.frame.mac, sample.manufacturer, DEF_LINELEN);
    sample.snr = mach_snr(s->device);
    sample.freq = mach_freq_get(s->device);
    sample.pps = 0;
    strings_clear(&sample.strings);
    wnet_add(s, sample.frame.mac, &sample, sizeof(sample));
    ui_info("New network %s", sample.frame.mac);
  }

  /* retrieve the entry from the array, and use it going forward */
  serial = wnet_exists(s, sample.frame.mac);
  assert(serial != -1);    
  wnet_get(s, serial, (void *)&network, &l);
  network->count.total++;
  pps_update_net(network);

  if(sample.frame.type == 0 && sample.frame.subtype == 8) {      /* beacon packet */      
    s->globalcount.beacon++;
    network->count.beacon++;
  } else if(sample.frame.type == 2 &&                            /* data packet (wep or cleartext) */
	    (sample.frame.subtype == 0 || 
	     sample.frame.subtype == 1 || 
	     sample.frame.subtype == 2 || 
	     sample.frame.subtype == 3)) {
    s->globalcount.data++;
    network->count.data++;      
    strings_add(&network->strings, pkt, pkthdr.len);
    if(s->capture == CAPTURE_DATA || network->capture)
      save_packet(s, &pkthdr, pkt);
  } else {                                                       /* something else */
    s->globalcount.martian++;
    network->count.martian++;
  }
    
  if(s->geiger)
    geiger_packet(s, &pkthdr, pkt);

  return 0;
}

int
ssid_sniff(ssidsniff_t *s) {
  fd_set fds;

  ui_init();
  counter_clear(&s->globalcount);
  signal(SIGINT, signal_handler);
  signal(SIGCONT, signal_handler);
  signal(SIGALRM, signal_handler);  
  alarm(s->delay);
  ui_redraw(s);

  while(s->running) {
    if(s->maxcount && s->globalcount.total >= s->maxcount)
      break;

    FD_ZERO(&fds);
    FD_SET(s->pcap_fd, &fds);
    FD_SET(0, &fds);
    
    if(select(s->pcap_fd+1, &fds, NULL, NULL, NULL) < 0) {
      if(errno == EINTR)
	continue;
      else
	break;
    }

    if(FD_ISSET(0, &fds)) {
      ui_process_events(s);
      continue;
    }

    if(!FD_ISSET(s->pcap_fd, &fds))
      continue;
    
    process_packet(s);
  }

  alarm(0);
  ui_finish();
  return 0;
}

int
default_writefile(char *p, int len, char *type) {
  struct tm *tm;
  char hostname[DEF_LINELEN];
  int serial = 0;
  struct stat st;
  time_t now;
  
  if(gethostname(hostname, DEF_LINELEN)) {
    strncpy(p, DEF_WRFILE, len);
    return 2;
  }

  now = time(NULL);
  if(!(tm = localtime(&now))) {
    strncpy(p, DEF_WRFILE, len);
    return 1;
  }
  
  while(serial++ < 99) {
    if(type)
      snprintf(p, len, "%s-%s-%04d-%02d-%02d-%02d.cap", hostname, type, tm->tm_year+1900, 
	       tm->tm_mon+1, tm->tm_mday, serial);
    else
      snprintf(p, len, "%s-%04d-%02d-%02d-%02d.cap", hostname, tm->tm_year+1900, 
	       tm->tm_mon+1, tm->tm_mday, serial);
    
    if(stat(p, &st))
      break;
  }
  
  return 0;
}

// XXX: support for freq maps and different freq bands
int
run_hopper(ssidsniff_t *s) {
  float hopping_seq[] = { 1, 6, 11, 3, 5, 7, 9, 2, 4, 8, 10, 12, 0 };
  int i;
  struct timespec t, t2;
  
  t.tv_sec = 1;
  t.tv_nsec = 0;
  i = 0;
  while(1) {
    if(!hopping_seq[i])
      i = 0;
    mach_channel_set(s->device, hopping_seq[i]);
    nanosleep(&t, &t2);
    i++;
  }
  return 0;
}

void
usage(char *argv0, int code) {
  /*               ------------------------------------------------------------------------------ */
  fprintf(stderr, "Usage: %s [-VdDg] [-i device] [-s snaplen] [-f filter] [-c maxcount]\n", argv0);
  fprintf(stderr, "       [-w file] [-v level]\n\n");
  fprintf(stderr, "\t-V              display version number and exit\n");
  fprintf(stderr, "\t-d              Save all data packets received\n");
  fprintf(stderr, "\t-D              Save all packets received\n");
  fprintf(stderr, "\t-i interface    Set the interface to listen on\n");
  fprintf(stderr, "\t-s snaplen      Maximum packet size to capture\n");
  fprintf(stderr, "\t-f filter       PCAP filter expression to apply\n");
  fprintf(stderr, "\t-c maxcount     Set maximum packets to read, then exit\n");
  fprintf(stderr, "\t-g              Geiger counter mode\n");
  fprintf(stderr, "\t-w file         Set PCAP capture file name\n");
  fprintf(stderr, "\t-v verbosity    set verbosity/debugging level\n");
  exit(code);
}

int
main(int argc, char **argv) {
  char c;
  int verbose;
  char *pdev = NULL;
  ssidsniff_t s;
  int pcap_snaplen;
  int pcap_promiscuous;
  char pcap_filter[DEF_LINELEN];
  struct bpf_program pcap_fcode;
  unsigned int pcap_if_net;
  unsigned int pcap_if_mask;
  char pcap_ebuf[PCAP_ERRBUF_SIZE];
  
  /* defaults */
  verbose = 1;
  s.maxcount = 0;
  s.device[0] = '\0';
  s.geiger = 0;
  s.dumpfile[0] = '\0';
  pcap_filter[0] = '\0';
  pcap_snaplen = 2000;
  pcap_promiscuous = 1;
  s.xpa = (xpaudio_t *)NULL;
  s.num_networks = 0;
  s.max_networks = DEF_WNET_ALLOC;
  s.start = time(NULL);
  s.delay = 1;
  s.running = 1;
  default_writefile(s.dumpfile, DEF_LINELEN, NULL);
  s.pd_dump = (pcap_dumper_t *)NULL;
  s.capture = CAPTURE_NONE;

  if(!(s.networks = (wnet_t *)xcalloc(s.max_networks, sizeof(wnet_t)))) {
    fprintf(stderr, "%s: Error allocating memory\n", argv[0]);
    exit(1);
  }

  while((c = getopt(argc, argv, "i:s:f:c:gw:v:VdD"))!=EOF) {
    switch(c) {
    case 'i':
      strncpy(s.device, optarg, PCAP_ERRBUF_SIZE);
      break;
    case 's':
      pcap_snaplen = atoi(optarg);
      break;
    case 'v':
      verbose = atoi(optarg);
      break;
    case 'f':
      strncpy(pcap_filter, optarg, DEF_LINELEN);
      break;
    case 'c':
      s.maxcount = atoi(optarg);
      break;
    case 'g':
      s.geiger = 1;
      break;
    case 'w':
      strncpy(s.dumpfile, optarg, DEF_LINELEN);
      break;
    case 'V':
      printf("%s %s\n", PACKAGE, VERSION);
      exit(0);
    case 'd':
      s.capture = CAPTURE_DATA;
      break;
    case 'D':
      s.capture = CAPTURE_ALL;
      break;
    case '?':
      usage(argv[0], 1);
      break;
    }
  }

  llog_init(LLOG_STDERR);
  llog_level(verbose);
    
  /* Find network device */
  if(!*s.device) {
    if(!(pdev = pcap_lookupdev(s.device))) {
      llog(0, "%s: No suitable device found\n", argv[0]);
      exit(1);
    }
    strncpy(s.device, pdev, PCAP_ERRBUF_SIZE);
    llog(1, "%s: Detected device %s\n", argv[0], s.device);
  } else {
    llog(1, "%s: Using device %s\n", argv[0], s.device);
  }

  /* set up interface */
  mach_set_if(s.device, 1);
  mach_set_monitor(s.device, 1);

  /* Initialize hopper */
#if 0
  if(!(s.hopper_pid = fork())) {
    run_hopper(&s);
    exit(0);
  }
#endif

  /* Initialize capturing */
  if(!(s.pd = pcap_open_live(s.device, pcap_snaplen, pcap_promiscuous, 100, pcap_ebuf))) {
    llog(0, "pcap_open_live(%s): %s\n", s.device, pcap_ebuf);
    return 2;
  }

  pcap_lookupnet(s.device, &pcap_if_net, &pcap_if_mask, pcap_ebuf);
  pcap_compile(s.pd, &pcap_fcode, pcap_filter, 1000, pcap_if_mask);
  pcap_setfilter(s.pd, &pcap_fcode);
  s.pcap_fd = pcap_fileno(s.pd);

  if(!mach_iswifi(s.device, NULL, 0)) {
    llog(1, "%s: Device %s isn't a wireless interface\n", argv[0], s.device);
    llog(1, "Continuing anyway\n");
    sleep(2);
  }
  
#ifdef HAVE_PCAP_SET_DATALINK
  pcap_set_datalink(s.pd, DLT_IEEE802_11);
#endif
  
  if(pcap_datalink(s.pd) != DLT_IEEE802_11) {
    llog(1, "%s: Link type %d (%s) isn't IEEE802_11, you probably need monitor mode\n", 	
	 argv[0], pcap_datalink(s.pd), 
#ifdef HAVE_PCAP_DATALINK_VAL_TO_NAME
	 pcap_datalink_val_to_name(pcap_datalink(s.pd))
#else
	 "Unknown"
#endif
	 );
    llog(1, "Continuing anyway\n");
    sleep(2);
  }
  
  llog(1, "%s: Starting capture with filter=%s on %s\n", argv[0], pcap_filter, s.device);
  gs = &s;
  ssid_sniff(&s);
  
  pcap_freecode(&pcap_fcode);
  pcap_close(s.pd);
  
  if(s.pd_dump)
     pcap_dump_close(s.pd_dump);
  
  if(s.xpa)
    xpaudio_destroy(s.xpa);
    
#if 0
  kill(s.hopper_pid, SIGTERM);
  wait(&verbose);
#endif
  llog_close();
  exit(0);
}

